import logging
from typing import Dict, List

from vmngclient.api.versions_utils import DeviceVersions, RepositoryAPI
from vmngclient.dataclasses import Device
from vmngclient.session import vManageSession

logger = logging.getLogger(__name__)


class PartitionManagerAPI:
    """
    API methods for partitions actions. All methods
    are exececutable on all device categories.
    """

    def __init__(self, session: vManageSession, device_versions: DeviceVersions, repository: RepositoryAPI) -> None:

        self.session = session
        self.device_versions = device_versions
        self.repository = repository

    def set_default_partition(self, devices: List[Device], version: str) -> str:
        """
        Method to set choosen software version as current version

        Args:
            version_to_default (str): software version to be set as default version

        Returns:
            str: action id
        """

        url = "/dataservice/device/action/defaultpartition"
        payload = {
            "action": "defaultpartition",
            "devices": self.device_versions.get_device_list_in_installed(version, devices),
            "deviceType": "vmanage",
        }
        set_default = dict(self.repository.session.post(url, json=payload).json())
        return set_default["id"]

    def remove_partition(self, devices: List[Device], version: str, force: bool = False) -> str:
        """
        Method to remove choosen software version from Vmanage repository

        Args:
            version (str): software version to be removed from repository

        Returns:
            str: action id
        """

        url = "/dataservice/device/action/removepartition"
        payload = {
            "action": "removepartition",
            "devices": self.device_versions.get_device_list_in_available(version, devices),
            "deviceType": "vmanage",
        }
        if force is False:
            invalid_devices = self._check_remove_partition_possibility(payload["devices"])
            if invalid_devices:
                raise ValueError(
                    f"Current or default version of devices with ids {invalid_devices} \
                        are equal to remove version. Action denied!"
                )
        remove_action: Dict[str, str] = self.repository.session.post(url, json=payload).json()
        return remove_action["id"]

    def _check_remove_partition_possibility(self, devices) -> List:

        devices_versions_repository = self.repository.get_devices_versions_repository(
            self.device_versions.device_category.value
        )
        invalid_devices = []
        for device in devices:

            if device["version"] in (
                devices_versions_repository[device["deviceId"]].current_version,
                devices_versions_repository[device["deviceId"]].default_version,
            ):
                invalid_devices.append((device["deviceId"]))
        return invalid_devices



import logging
from enum import Enum
from pathlib import PurePath
from typing import Any, Dict, List

from attr import define  # type: ignore
from clint.textui.progress import Bar as ProgressBar  # type: ignore
from requests_toolbelt.multipart.encoder import MultipartEncoder, MultipartEncoderMonitor  # type: ignore

from vmngclient.api.versions_utils import DeviceVersions, RepositoryAPI
from vmngclient.dataclasses import Device
from vmngclient.session import vManageSession

logger = logging.getLogger(__name__)


class Family(Enum):
    VEDGE = "vedge"
    VMANAGE = "vmanage"


class VersionType(Enum):
    VMANAGE = "vmanage"


class DeviceType(Enum):
    CONTROLLER = "controller"
    VEDGE = "vedge"
    VMANAGE = "vmanage"


class DeviceClass(Enum):
    VEDGE = "vedge"
    VMANAGE = "vmanage"
    VSMART = "vsmart"
    VBOND = "vbond"
    CEDGE = "cedge"


@define
class InstallSpecification:

    family: Family
    version_type: VersionType
    device_type: DeviceType
    device_class: DeviceClass


class InstallSpecHelper(Enum):
    VMANAGE = InstallSpecification(Family.VMANAGE, VersionType.VMANAGE, DeviceType.VMANAGE, DeviceClass.VMANAGE)
    VSMART = InstallSpecification(Family.VEDGE, VersionType.VMANAGE, DeviceType.CONTROLLER, DeviceClass.VSMART)
    VBOND = InstallSpecification(Family.VEDGE, VersionType.VMANAGE, DeviceType.CONTROLLER, DeviceClass.VBOND)
    VEDGE = InstallSpecification(Family.VEDGE, VersionType.VMANAGE, DeviceType.VEDGE, DeviceClass.VEDGE)
    CEDGE = InstallSpecification(Family.VEDGE, VersionType.VMANAGE, DeviceType.VEDGE, DeviceClass.CEDGE)


class SoftwareActionAPI:
    """
    API methods for software actions. All methods
    are exececutable on all device categories.
    """

    def __init__(self, session: vManageSession, device_versions: DeviceVersions, repository: RepositoryAPI) -> None:

        self.session = session
        self.device_versions = device_versions
        self.repository = repository

    def activate_software(self, version_to_activate: str, devices: List[Device]) -> str:
        """
        Method to set choosen version as current version

        Args:
            version_to_activate (str): version to be set as current version

        Returns:
            str: action id
        """

        url = "/dataservice/device/action/changepartition"
        payload = {
            "action": "changepartition",
            "devices": self.device_versions.get_device_list_in_available(version_to_activate, devices),
            "deviceType": "vmanage",
        }
        activate = dict(self.session.post(url, json=payload).json())
        return activate["id"]

    def upgrade_software(
        self,
        devices: List[Device],
        software_image: str,
        install_spec: InstallSpecification,
        reboot: bool,
        sync: bool = True,
    ) -> str:
        """
        Method to install new software

        Args:
            software_image (str): path to software image
            install_spec (InstallSpecification): specification of devices
            on which the action is to be performed
            reboot (bool): reboot device after action end
            sync (bool, optional): Synchronize settings. Defaults to True.

        Raises:
            ValueError: Raise error if downgrade in certain cases

        Returns:
            str: action id
        """

        url = "/dataservice/device/action/install"
        payload: Dict[str, Any] = {
            "action": "install",
            "input": {
                "vEdgeVPN": 0,
                "vSmartVPN": 0,
                "family": install_spec.family.value,
                "version": self.repository.get_image_version(software_image),
                "versionType": install_spec.version_type.value,
                "reboot": reboot,
                "sync": sync,
            },
            "devices": self.device_versions.get_device_list(devices),
            "deviceType": install_spec.device_type.value,
        }
        if install_spec.family.value in (DeviceClass.VMANAGE.value, DeviceClass.CEDGE.value):
            incorrect_devices = self._downgrade_check(
                payload["devices"],
                payload["input"]["version"],
                install_spec.family.value,
            )
            if incorrect_devices:
                raise ValueError(
                    f"Current version of devices with id's {incorrect_devices} is \
                    higher than upgrade version. Action denied!"
                )
        upgrade = dict(self.session.post(url, json=payload).json())
        return upgrade["id"]

    def _create_callback(self, encoder: MultipartEncoder):

        bar = ProgressBar(expected_size=encoder._calculate_length(), filled_char="=")

        def callback(monitor: MultipartEncoderMonitor):
            bar.show(monitor.bytes_read)

        return callback

    def upload_image(self, image_path: str) -> int:
        """
        Upload software image 'tar.gz' to Vmanage
        software repository

        Args:
            image_path (str): path to software image

        Returns:
            str: Response status code
        """
        url = "/dataservice/device/action/software/package"
        encoder = MultipartEncoder(
            fields={"file": (PurePath(image_path).name, open(image_path, "rb"), "application/x-gzip")}
        )
        callback = self._create_callback(encoder)
        monitor = MultipartEncoderMonitor(encoder, callback)
        upload = self.session.post(url, data=monitor, headers={"content-type": monitor.content_type})
        return upload.status_code

    def _downgrade_check(self, devices, version_to_upgrade: str, family) -> List:
        """
        Check if upgrade operation is not actually a downgrade opeartion.
        If so, in some cases action is being blocked.

        Args:
            version_to_upgrade (str): version to upgrade
            devices_category (DeviceCategory): devices category

        Returns:
            Union[None, List]: [None, list of devices with no permission to downgrade]
        """
        incorrect_devices = []
        devices_versions_repo = self.repository.get_devices_versions_repository(
            self.device_versions.device_category.value
        )
        for dev in devices:
            dev_current_version = str(devices_versions_repo[dev["deviceId"]].current_version)
            splited_version_to_upgrade = version_to_upgrade.split(".")
            for priority, label in enumerate(dev_current_version.split("-")[0].split(".")):
                if str(label) > str(splited_version_to_upgrade[priority]):
                    if family == "vmanage" and label == 2:
                        continue
                    incorrect_devices.append(dev["deviceId"])
                    break
                elif str(label) < str(splited_version_to_upgrade[priority]):
                    break
        return incorrect_devices



import unittest
from unittest.mock import Mock, patch

from vmngclient.api.software_action_api import Family, InstallSpecHelper, SoftwareActionAPI
from vmngclient.api.versions_utils import DeviceCategory, DeviceSoftwareRepository, DeviceVersions, RepositoryAPI
from vmngclient.dataclasses import Device


class TestSoftwareAcionAPI(unittest.TestCase):
    def setUp(self):
        self.device = Device(
            personality="vedge",
            uuid="mock_uuid",
            id="mock_ip",
            hostname="mock_host",
            reachability="reachable",
            local_system_ip="mock_ip",
            status="normal",
            memUsage=1.0,
            connected_vManages=["192.168.0.1"],
            model="vedge-cloud",
        )
        self.DeviceSoftwareRepository_obj = {
            "mock_uuid": DeviceSoftwareRepository(
                ["ver1", "ver2", "curr_ver"],
                ["ver1", "ver2"],
                "20.8",
                "def_ver",
                "mock_uuid",
            ),
        }

        self.mock_devices = [{"deviceId": "mock_uuid", "deviceIP": "mock_ip", "version": "ver1"}]
        self.install_spec = InstallSpecHelper.VMANAGE.value

        mock_session = Mock()
        self.mock_repository_object = RepositoryAPI(mock_session)
        self.mock_device_versions = DeviceVersions(self.mock_repository_object, DeviceCategory.CONTROLLERS)
        self.mock_software_action_obj = SoftwareActionAPI(
            mock_session, self.mock_device_versions, self.mock_repository_object
        )

    @patch.object(SoftwareActionAPI, "_downgrade_check")
    @patch.object(RepositoryAPI, "get_image_version")
    def test_upgrade_software_if_downgrade_check_not_none(self, mock_get_image_version, mock_downgrade_check):

        # Prepare mock data
        mock_downgrade_check.return_value = ["mock_uuid"]

        # Assert
        self.assertRaises(
            ValueError,
            self.mock_software_action_obj.upgrade_software,
            [self.device],
            "path",
            self.install_spec,
            True,
            True,
        )

    @patch("vmngclient.session.vManageSession")
    @patch.object(SoftwareActionAPI, "_downgrade_check")
    @patch.object(RepositoryAPI, "get_image_version")
    def test_upgrade_software_if_downgrade_check_is_none(
        self, mock_get_image_version, mock_downgrade_check, mock_session
    ):

        # Prepare mock data
        mock_downgrade_check.return_value = False
        self.mock_repository_object.session.post.return_value.json.return_value = {"id": "mock_action_id"}
        mock_session.post.return_value = {"id": "mock_action_id"}
        # self.mock_software_action_obj.session.post.json.return_value = {"id": "mock_action_id"}

        # Assert
        answer = self.mock_software_action_obj.upgrade_software([self.device], "path", self.install_spec, True, True)
        self.assertEqual(answer, "mock_action_id", "action ids not equal")

    @patch.object(RepositoryAPI, "get_devices_versions_repository")
    def test_downgrade_check_no_incorrect_devices(self, mock_get_devices_versions_repository):
        # Preapre mock data
        upgrade_version = "20.9"
        mock_get_devices_versions_repository.return_value = self.DeviceSoftwareRepository_obj
        mock_devices = [{"deviceId": "mock_uuid", "deviceIP": "mock_ip", "version": upgrade_version}]
        self.mock_repository_object.devices = mock_devices
        # Assert
        answer = self.mock_software_action_obj._downgrade_check(mock_devices, upgrade_version, Family.VMANAGE.value)
        self.assertEqual(answer, [], "downgrade detected, but should not")

    @patch.object(RepositoryAPI, "get_devices_versions_repository")
    def test_downgrade_check_incorrect_devices_exists(self, mock_get_devices_versions_repository):
        # Preapre mock data
        upgrade_version = "20.6"
        mock_get_devices_versions_repository.return_value = self.DeviceSoftwareRepository_obj
        mock_devices = [{"deviceId": "mock_uuid", "deviceIP": "mock_ip", "version": upgrade_version}]
        self.mock_repository_object.devices = mock_devices
        # Assert
        answer = self.mock_software_action_obj._downgrade_check(mock_devices, upgrade_version, Family.VMANAGE.value)
        self.assertEqual(answer, ["mock_uuid"], "downgrade detected, but should not")



import unittest
from unittest.mock import Mock, patch

from vmngclient.api.partition_manager_api import PartitionManagerAPI
from vmngclient.api.versions_utils import DeviceCategory, DeviceSoftwareRepository, DeviceVersions, RepositoryAPI
from vmngclient.dataclasses import Device


class TestPartitionManagerAPI(unittest.TestCase):
    def setUp(self):
        self.device = Device(
            personality="vedge",
            uuid="mock_uuid",
            id="mock_ip",
            hostname="mock_host",
            reachability="reachable",
            local_system_ip="mock_ip",
            status="normal",
            memUsage=1.0,
            connected_vManages=["192.168.0.1"],
            model="vedge-cloud",
        )
        self.DeviceSoftwareRepository_obj = {
            "mock_uuid": DeviceSoftwareRepository(
                ["ver1", "ver2", "curr_ver"],
                ["ver1", "ver2"],
                "curr_ver",
                "def_ver",
                "mock_uuid",
            ),
        }

        self.mock_devices = [{"deviceId": "mock_uuid", "deviceIP": "mock_ip", "version": "ver1"}]
        mock_session = Mock()
        self.mock_repository_object = RepositoryAPI(mock_session)
        self.mock_device_versions = DeviceVersions(self.mock_repository_object, DeviceCategory.CONTROLLERS)
        self.mock_partition_manager_obj = PartitionManagerAPI(
            mock_session, self.mock_device_versions, self.mock_repository_object
        )

    @patch.object(PartitionManagerAPI, "_check_remove_partition_possibility")
    @patch.object(DeviceVersions, "get_device_list_in_available")
    def test_remove_partition_raise_error_force_false(self, mock_complete_device_list, mock_check_remove):

        # Prepare mock data
        mock_complete_device_list.return_value = Mock()
        self.mock_device_versions.get_device_list_in_available(
            "ver1", "available_versions"
        ).return_value = self.mock_devices
        mock_check_remove.return_value = ["mock_uuid"]
        # self.mock_repository_object.devices = self.mock_devices

        # Assert
        self.assertRaises(ValueError, self.mock_partition_manager_obj.remove_partition, [self.device], "ver1", False)

    @patch.object(DeviceVersions, "get_device_list_in_available")
    def test_remove_partition_if_force_true(self, mock_get_device_list):

        # Prepare mock data
        mock_get_device_list.return_value = Mock()
        mock_devices = Mock()
        mock_devices.return_value = self.mock_devices
        mock_get_device_list.return_value = self.mock_devices
        self.mock_repository_object.devices = self.mock_devices
        self.mock_repository_object.session.post.return_value.json.return_value = {"id": "mock_action_id"}

        # Assert
        answer = self.mock_partition_manager_obj.remove_partition([self.device], "ver1", True)
        self.assertEqual(answer, "mock_action_id", "action ids not equal")

    @patch.object(PartitionManagerAPI, "_check_remove_partition_possibility")
    @patch.object(DeviceVersions, "get_device_list_in_available")
    def test_remove_partition_not_raise_error_force_false(self, mock_get_device_list, mock_check_remove):

        # Prepare mock data
        mock_get_device_list.return_value = Mock()
        self.mock_device_versions.get_device_list_in_available(
            "ver1", "available_versions"
        ).return_value = self.mock_devices
        mock_check_remove.return_value = []
        self.mock_repository_object.devices = self.mock_devices
        self.mock_repository_object.session.post.return_value.json.return_value = {"id": "mock_action_id"}

        # Assert
        answer = self.mock_partition_manager_obj.remove_partition("ver1", True)
        self.assertEqual(answer, "mock_action_id", "action ids not equal")

    @patch.object(RepositoryAPI, "get_devices_versions_repository")
    def test_check_remove_partition_possibility_if_version_incorrect(self, mock_get_devices_versions_repository):

        # Prepare mock data
        mock_get_devices_versions_repository.return_value = self.DeviceSoftwareRepository_obj
        mock_devices = [{"deviceId": "mock_uuid", "deviceIP": "mock_ip", "version": "curr_ver"}]

        # Assert
        answer = self.mock_partition_manager_obj._check_remove_partition_possibility(mock_devices)
        self.assertEqual(answer, ["mock_uuid"], "lists are not equal")

    @patch.object(RepositoryAPI, "get_devices_versions_repository")
    def test_check_remove_partition_possibility_if_version_correct(self, mock_get_devices_versions_repository):

        # Prepare mock data
        mock_get_devices_versions_repository.return_value = self.DeviceSoftwareRepository_obj
        mock_devices = [{"deviceId": "mock_uuid", "deviceIP": "mock_ip", "version": "any_ver"}]

        # Assert
        answer = self.mock_partition_manager_obj._check_remove_partition_possibility(mock_devices)
        self.assertEqual(answer, [], "lists are not equal")



import unittest
from unittest.mock import MagicMock, Mock, patch

from ciscoconfparse import CiscoConfParse  # type: ignore

from vmngclient.api.templates import AttachedError, NameAlreadyExistError, NotFoundError, TemplateAPI
from vmngclient.dataclasses import Device, Template
from vmngclient.utils.creation_tools import create_dataclass
from vmngclient.utils.operation_status import OperationStatus
from vmngclient.utils.personality import Personality
from vmngclient.utils.reachability import Reachability


class TestTemplateAPI(unittest.TestCase):
    def setUp(self):
        self.data_template = [
            {
                "deviceType": "vedge-C8000V",
                "lastUpdatedBy": "user",
                "resourceGroup": "global",
                "templateClass": "cedge",
                "configType": "file",
                "templateId": "dummy_id_1",
                "factoryDefault": False,
                "templateName": "template_1",
                "devicesAttached": 0,
                "templateDescription": "dummy template 1",
                "draftMode": "Disabled",
                "lastUpdatedOn": 0,
                "templateAttached": 0,
            },
            {
                "deviceType": "vedge-cloud",
                "lastUpdatedBy": "admin",
                "resourceGroup": "global",
                "templateClass": "vedge",
                "configType": "file",
                "templateId": "dummy_id_2",
                "factoryDefault": False,
                "templateName": "template_2",
                "devicesAttached": 1,
                "templateDescription": "dummy template 2",
                "draftMode": "Disabled",
                "lastUpdatedOn": 0,
                "templateAttached": 0,
            },
        ]
        self.templates = [create_dataclass(Template, template) for template in self.data_template]
        self.device_info = Device(
            personality=Personality.EDGE,
            uuid="dummy_uuid",
            id="162.168.0.1",
            hostname="dummy_host",
            reachability=Reachability.REACHABLE,
            local_system_ip="192.168.0.1",
            memUsage=1.0,
            connected_vManages=["192.168.0.1"],
            model="vedge-cloud",
            status="normal",
        )

    @patch("vmngclient.session.vManageSession")
    def test_templates_success(self, mock_session):

        # Arrange
        mock_session.get_data.return_value = self.data_template

        # Act
        answer = TemplateAPI(mock_session).templates

        # Assert
        self.assertEqual(answer, self.templates)

    @patch("vmngclient.session.vManageSession")
    def test_templates_no_data(self, mock_session):

        # Arrange
        mock_session.get_data.return_value = []

        # Act
        answer = TemplateAPI(mock_session).templates

        # Assert
        self.assertEqual(answer, [])

    @patch.object(TemplateAPI, "templates")
    def test_get_success(self, mock_templates):

        # Arrage
        MockTemplates = Mock()
        mock_templates.return_value = MockTemplates
        session = Mock()
        test_object = TemplateAPI(session)
        test_object.templates = self.templates

        # Act
        answer = [test_object.get(template.name) for template in self.templates]

        # Assert
        self.assertEqual(answer, self.templates)

    @patch.object(TemplateAPI, "templates")
    def test_get_exception(self, mock_templates):

        # Arrage
        MockTemplates = Mock()
        mock_templates.return_value = MockTemplates
        session = Mock()
        test_object = TemplateAPI(session)
        test_object.templates = self.templates

        # Act
        def answer():
            return test_object.get("no_exist_name")

        # Assert
        self.assertRaises(NotFoundError, answer)

    @patch.object(TemplateAPI, "templates")
    def test_get_id_success(self, mock_templates):

        # Arrage
        MockTemplates = Mock()
        mock_templates.return_value = MockTemplates
        session = Mock()
        test_object = TemplateAPI(session)
        test_object.templates = self.templates

        # Act
        answer = [test_object.get_id(template.name) for template in self.templates]

        ids = [template.id for template in self.templates]
        # Assert
        self.assertEqual(answer, ids)

    @patch.object(TemplateAPI, "templates")
    def test_get_id_exception(self, mock_templates):

        # Arrage
        MockTemplates = Mock()
        mock_templates.return_value = MockTemplates
        session = Mock()
        test_object = TemplateAPI(session)
        test_object.templates = self.templates

        # Act
        def answer():
            return test_object.get_id("no_exist_name")

        # Assert
        self.assertRaises(NotFoundError, answer)

    @patch.object(TemplateAPI, "get_operation_status")
    def test_wait_for_complete_true(self, mock_operation):

        # Arrage
        MockOperatrion = Mock()
        mock_operation.return_value = MockOperatrion
        session = Mock()
        test_object = TemplateAPI(session)
        test_object.get_operation_status.return_value = [
            OperationStatus.SUCCESS,
            OperationStatus.SUCCESS,
        ]
        operation_id = "operation_id"

        # Act
        answer = test_object.wait_for_complete(operation_id)

        # Assert
        self.assertTrue(answer)

    @patch.object(TemplateAPI, "get_operation_status")
    def test_wait_for_complete_false(self, mock_operation):

        # Arrage
        MockOperatrion = Mock()
        mock_operation.return_value = MockOperatrion
        session = Mock()
        test_object = TemplateAPI(session)
        test_object.get_operation_status.return_value = [
            OperationStatus.IN_PROGRESS,
            OperationStatus.SUCCESS,
        ]
        operation_id = "operation_id"

        # Act
        answer = test_object.wait_for_complete(operation_id, timeout_seconds=2, sleep_seconds=1)

        # Assert
        self.assertFalse(answer)

    @patch("vmngclient.session.vManageSession")
    @patch.object(TemplateAPI, "templates")
    @patch.object(TemplateAPI, "wait_for_complete")
    def test_attach_exist_template(self, mock_session, mock_wait_for_complete, mock_templates):

        # Arrage
        mock_session.post_json.return_value = {"id": "operation_id"}
        test_object = TemplateAPI(mock_session)

        # mock templates
        MockTemplates = Mock()
        mock_templates.return_value = MockTemplates
        test_object.templates = self.templates

        # mock wait complete
        MockWaitComplete = Mock()
        mock_wait_for_complete.return_value = MockWaitComplete
        test_object.wait_for_complete.return_value = True

        # Act
        answer = test_object.attach("template_1", self.device_info)

        # Assert
        self.assertTrue(answer)

    @patch("vmngclient.session.vManageSession")
    @patch.object(TemplateAPI, "templates")
    @patch.object(TemplateAPI, "wait_for_complete")
    def test_attach_no_exist_template(self, mock_session, mock_wait_for_complete, mock_templates):

        # Arrage
        mock_session.post_json.return_value = {"id": "operation_id"}
        test_object = TemplateAPI(mock_session)

        # mock templates
        MockTemplates = Mock()
        mock_templates.return_value = MockTemplates
        test_object.templates = self.templates

        # mock wait complete
        MockWaitComplete = Mock()
        mock_wait_for_complete.return_value = MockWaitComplete
        test_object.wait_for_complete.return_value = True

        # Act
        answer = test_object.attach("no_exist_template", self.device_info)

        # Assert
        self.assertFalse(answer)

    @patch("vmngclient.session.vManageSession")
    @patch.object(TemplateAPI, "wait_for_complete")
    def test_device_to_cli_true(self, mock_session, mock_wait_for_complete):

        # Arrage
        mock_session.post_json.return_value = {"id": "operation_id"}
        test_object = TemplateAPI(mock_session)

        # mock wait complete
        MockWaitComplete = Mock()
        mock_wait_for_complete.return_value = MockWaitComplete
        test_object.wait_for_complete.return_value = True

        # Act
        answer = test_object.device_to_cli(self.device_info)

        # Assert
        self.assertTrue(answer)

    @patch("vmngclient.session.vManageSession")
    def test_get_operatrion_status_success(self, mock_session):

        # Arrage
        mock_session.get_data.return_value = [
            {"status": "In progress"},
            {"status": "Failure"},
            {"status": "Success"},
            {"status": "Scheduled"},
        ]
        test_object = TemplateAPI(mock_session)
        operation_id = "dummy_id"

        # Act
        answer = test_object.get_operation_status(operation_id)

        # Assert
        self.assertEqual(
            answer,
            [
                OperationStatus.IN_PROGRESS,
                OperationStatus.FAILURE,
                OperationStatus.SUCCESS,
                OperationStatus.SCHEDULED,
            ],
        )

    @patch("vmngclient.session.vManageSession")
    def test_get_operatrion_status_no_data(self, mock_session):

        # Arrage
        mock_session.get_data.return_value = []
        test_object = TemplateAPI(mock_session)
        operation_id = "dummy_id"

        # Act
        answer = test_object.get_operation_status(operation_id)

        # Assert
        self.assertEqual(answer, [])

    @patch("vmngclient.session.vManageSession")
    @patch.object(TemplateAPI, "templates")
    def test_delete_success(self, mock_session, mock_templates):

        # Arrage
        MockResponse = MagicMock()
        MockResponse.status_code = 200
        mock_session.delete.return_value = MockResponse
        test_object = TemplateAPI(mock_session)

        # mock templates
        MockTemplates = Mock()
        mock_templates.return_value = MockTemplates
        test_object.templates = self.templates

        # Act
        answer = test_object.delete("template_1")
        # Assert
        self.assertTrue(answer)

    @patch("vmngclient.session.vManageSession")
    @patch.object(TemplateAPI, "templates")
    def test_delete_wrong_status(self, mock_session, mock_templates):

        # Arrage
        MockResponse = MagicMock()
        MockResponse.ok = False
        mock_session.delete.return_value = MockResponse
        test_object = TemplateAPI(mock_session)

        # mock templates
        MockTemplates = Mock()
        mock_templates.return_value = MockTemplates
        test_object.templates = self.templates

        # Act
        answer = test_object.delete("template_1")

        # Assert
        self.assertFalse(answer)

    @patch("vmngclient.session.vManageSession")
    @patch.object(TemplateAPI, "templates")
    def test_delete_exception(self, mock_session, mock_templates):

        # Arrage
        MockResponse = MagicMock()
        MockResponse.status = 404
        mock_session.delete.return_value = MockResponse
        test_object = TemplateAPI(mock_session)

        # mock templates
        MockTemplates = Mock()
        mock_templates.return_value = MockTemplates
        test_object.templates = self.templates

        # Act
        def answer():
            return test_object.delete("template_2")

        # Assert
        self.assertRaises(AttachedError, answer)

    @patch("vmngclient.session.vManageSession")
    @patch.object(TemplateAPI, "templates")
    def test_create_exception(self, mock_session, mock_templates):

        # Arrage
        test_object = TemplateAPI(mock_session)

        # mock templates
        MockTemplates = Mock()
        mock_templates.return_value = MockTemplates
        test_object.templates = self.templates
        config = CiscoConfParse([])

        # Act
        def answer():
            return test_object.create(self.device_info, "template_1", "new_description", config)

        # Assert
        self.assertRaises(NameAlreadyExistError, answer)



import logging
from typing import List, cast

from attr import define, field  # type: ignore
from tenacity import retry, retry_if_result, stop_after_attempt, wait_fixed  # type: ignore

from vmngclient.session import vManageSession
from vmngclient.utils.creation_tools import FIELD_NAME, create_dataclass
from vmngclient.utils.operation_status import OperationStatus, OperationStatusId

logger = logging.getLogger(__name__)


@define
class TaskStatus:
    status: str
    status_id: str = field(metadata={FIELD_NAME: "statusId"})
    activity: List[str]


def wait_for_completed(
    session: vManageSession,
    action_id: str,
    timeout_seconds: int = 300,
    sleep_seconds: int = 5,
    exit_statuses: List[OperationStatus] = [
        OperationStatus.SUCCESS,
        OperationStatus.FAILURE,
    ],
    exit_statuses_ids: List[OperationStatusId] = [
        OperationStatusId.SUCCESS,
        OperationStatusId.FAILURE,
    ],
    activity_text: str = "",
) -> TaskStatus:
    """
    Method to check action status

    Example:
        session = create_vManageSession(ip_address,admin_username,password,port=port)
        devices = DevicesAPI(session).devices
        vsmart_device = [dev for dev in devices if dev.personality == Personality.VSMART][0]

        reboot_action = RebootAction(session,devices)
        reboot_action.execute()

        # Keep asking for reboot status until it's not in exit_statuses (Failure or Success)
          or timeout is not achieved (3000s)
        task = wait_for_completed(session,reboot_action.action_id,3000)
        if task.status == OperationStatus.SUCCESS.value:
            #do something
        else:
            #do something else

    Args:
        session (vManageSession): session
        action_id (str): inspected action id
        timeout_seconds (int): After this time, function will stop requesting action status
        sleep_seconds (int): interval between action status requests
        exit_statuses (Union[List[OperationStatus], str]): actions statuses that cause stop requesting action status
        exit_statuses_ids (Union[List[OperationStatusId], str]): actions statuses ids
            that cause stop requesting action status id
        activity_text (str): activity text

    Returns:
        task (TaskStatus):
    """
    action_url = "/dataservice/device/action/status/"
    exit_statuses = [cast(OperationStatus, exit_status.value) for exit_status in exit_statuses]
    exit_statuses_ids = [cast(OperationStatusId, exit_status_id.value) for exit_status_id in exit_statuses_ids]

    def check_status(task: TaskStatus) -> bool:
        """
        Function checks if condition is met. If so,
        wait_for_completed stops asking for task status

        Args:
            status (str): status of task
            status_id (str): status id of task
            activity (str): activity text

        Returns:
            bool: False if condition is met
        """

        if (task.status in exit_statuses) and (task.status_id in exit_statuses_ids):
            if not activity_text or activity_text in task.activity:
                return False
        return True

    def log_exception(self) -> None:
        logger.error("Operation status not achieved in given time")

    @retry(
        wait=wait_fixed(sleep_seconds),
        stop=stop_after_attempt(int(timeout_seconds / sleep_seconds)),
        retry=retry_if_result(check_status),
        retry_error_callback=log_exception,
    )
    def wait_for_action_finish() -> TaskStatus:
        """
        Keep asking for task status, status_id,
        activity(optional), utill check_status is True

        Returns:
            TaskStatus: TaskStatus instance
        """
        url = f"{action_url}{action_id}"
        action_data = session.get_data(url)[0]
        task = create_dataclass(TaskStatus, action_data)
        logger.debug(
            f"Statuses of action {action_id} is: "
            f"status: {task.status}, status_id: {task.status_id}, activity: {task.activity}."
        )
        return task

    return wait_for_action_finish()



import unittest
from unittest.mock import Mock, patch

from vmngclient.api.versions_utils import DeviceCategory, DeviceSoftwareRepository, DeviceVersions, RepositoryAPI
from vmngclient.dataclasses import Device


class TestRepositoryAPI(unittest.TestCase):
    def setUp(self):
        self.device = Device(
            personality="vedge",
            uuid="mock_uuid",
            id="mock_ip",
            hostname="mock_host",
            reachability="reachable",
            local_system_ip="mock_ip",
            status="normal",
            memUsage=1.0,
            connected_vManages=["192.168.0.1"],
            model="vedge-cloud",
        )
        self.DeviceSoftwareRepository_obj = {
            "mock_uuid": DeviceSoftwareRepository(
                ["ver1", "ver2", "curr_ver"],
                ["ver1", "ver2"],
                "curr_ver",
                "def_ver",
                "mock_uuid",
            )
        }

    @patch("vmngclient.session.Session")
    def test_get_image_version_if_image_available(self, mock_session):

        versions_response = [{"availableFiles": "vmanage-20.9.1-x86_64.tar.gz", "versionName": "20.9.1"}]
        mock_session.get_data.return_value = versions_response
        image_version = "20.9.1"
        answer = RepositoryAPI(mock_session).get_image_version("vmanage-20.9.1-x86_64.tar.gz")

        self.assertEqual(answer, image_version, "not same version")

    @patch("vmngclient.session.Session")
    def test_get_image_version_if_image_unavailable(self, mock_session):

        api_mock_response = [{"availableFiles": "vmanage-20.9.2-x86_64.tar.gz", "versionName": "20.9.1"}]
        mock_session.get_data.return_value = api_mock_response
        image_version = None
        answer = RepositoryAPI(mock_session).get_image_version("vmanage-20.9.1-x86_64.tar.gz")

        self.assertEqual(answer, image_version, "not same version")

    @patch("vmngclient.session.Session")
    def test_get_devices_versions_repository(self, mock_session):

        api_mock_response = [
            {
                "availableVersions": ["ver1", "ver2"],
                "version": "curr_ver",
                "defaultVersion": "def_ver",
                "uuid": "mock_uuid",
            }
        ]
        mock_session.get_data.return_value = api_mock_response
        mock_repository_object = RepositoryAPI(mock_session)

        answer = mock_repository_object.get_devices_versions_repository(DeviceCategory.CONTROLLERS.value)

        self.assertEqual(
            answer["mock_uuid"],
            self.DeviceSoftwareRepository_obj["mock_uuid"],
            "DeviceSoftwareRepository object created uncorrectly",
        )

    @patch.object(RepositoryAPI, "get_devices_versions_repository")
    def test_get_device_list_in_available(self, mock_get_devices_versions_repository):
        # Prepare mock data
        mock_get_devices_versions_repository.return_value = Mock()
        mock_session = Mock()
        mock_repository_object = RepositoryAPI(mock_session)
        mock_device_versions = DeviceVersions(mock_repository_object, DeviceCategory.CONTROLLERS.value)
        mock_get_devices_versions_repository.return_value = self.DeviceSoftwareRepository_obj
        answer = mock_device_versions.get_device_list_in_available("ver1", [self.device])
        expected_result = [{"deviceId": "mock_uuid", "deviceIP": "mock_ip", "version": "ver1"}]

        # Assertcom
        self.assertEqual(
            answer,
            expected_result,
            "Version add incorrectly",
        )

    @patch.object(RepositoryAPI, "get_devices_versions_repository")
    def test_get_device_list_if_in_installed(self, mock_get_devices_versions_repository):
        # Prepare mock data
        mock_get_devices_versions_repository.return_value = Mock()
        mock_session = Mock()
        mock_repository_object = RepositoryAPI(mock_session)
        mock_device_versions = DeviceVersions(mock_repository_object, DeviceCategory.CONTROLLERS.value)
        mock_get_devices_versions_repository.return_value = self.DeviceSoftwareRepository_obj
        answer = mock_device_versions.get_device_list_in_installed("ver1", [self.device])
        expected_result = [{"deviceId": "mock_uuid", "deviceIP": "mock_ip", "version": "ver1"}]

        # Assert
        self.assertEqual(
            answer,
            expected_result,
            "Version add incorrectly",
        )



import logging
from enum import Enum
from pathlib import PurePath
from typing import Dict, List, Union

from attr import define, field  # type: ignore

from vmngclient.dataclasses import Device
from vmngclient.session import vManageSession
from vmngclient.utils.creation_tools import FIELD_NAME, create_dataclass

logger = logging.getLogger(__name__)


class DeviceCategory(Enum):
    CONTROLLERS = "controllers"
    VEDGES = "vedges"


@define
class DeviceSoftwareRepository:
    installed_versions: List[str] = field(default=None)
    available_versions: List[str] = field(default=None, metadata={FIELD_NAME: "availableVersions"})
    current_version: str = field(default=None, metadata={FIELD_NAME: "version"})
    default_version: str = field(default=None, metadata={FIELD_NAME: "defaultVersion"})
    device_id: str = field(default=None, metadata={FIELD_NAME: "uuid"})


class RepositoryAPI:
    """
    API methods to get information about images and devices versions
    """

    def __init__(
        self,
        session: vManageSession,
    ):
        self.session = session

    def get_all_software_images(self) -> list:
        """
        Method to get all info about all software images stored
        in Vmanage repository

        Returns:
            list: software images list
        """
        url = "/dataservice/device/action/software/images?imageType=software"
        software_images = list(self.session.get_data(url))
        return software_images

    def get_devices_versions_repository(self, device_category) -> Dict[str, DeviceSoftwareRepository]:
        """
        Method for create DeviceSoftwareRepository dataclass,
        which cointains information about all possible version types for certain devices

        Returns:
            Dict[str, DeviceSoftwareRepository]: Dictionary containing all versions
            information
        """

        url = f"/dataservice/system/device/{device_category}"
        devices_versions_info = self.session.get_data(url)
        devices_versions_repository = {}
        for device in devices_versions_info:
            device_all_versions = create_dataclass(DeviceSoftwareRepository, device)
            device_all_versions.installed_versions = [version for version in device_all_versions.available_versions]
            device_all_versions.installed_versions.append(device_all_versions.current_version)
            devices_versions_repository[device_all_versions.device_id] = device_all_versions
        return devices_versions_repository

    def get_image_version(self, software_image: str) -> Union[str, None]:
        """
        Method to get proper software image version

        Args:
            software_image (str): path to software image

        Returns:
            Union[str, None]: image version or None
        """

        image_name = PurePath(software_image).name
        software_images = self.get_all_software_images()
        for img in software_images:
            if image_name in img["availableFiles"]:
                image_version = img["versionName"]
                return image_version
        logger.error(f"Software image {image_name} is not in available images")
        return None


class DeviceVersions:
    """
    Methods to prepare devices list for payload
    """

    def __init__(self, repository: RepositoryAPI, device_category: DeviceCategory):
        self.repository = repository
        self.device_category = device_category

    def _get_device_list_in(self, version_to_set_up: str, devices: List[Device], version_type: str) -> List[dict]:
        """
        Create version key for every device dict in device list, if requested version
        is in requested version type

        Args:
            version_to_set_up (str): requested version
            devices List[Device]: list of Device dataclass instances
            version_type: type of version (installed, available, etc.)

        Returns:
            list : list of devices
        """
        devs = [{"deviceId": dev.uuid, "deviceIP": dev.id} for dev in devices]
        all_dev_versions = self.repository.get_devices_versions_repository(self.device_category)
        for dev in devs:
            dev_installed_versions = getattr(all_dev_versions[dev["deviceId"]], version_type)
            for version in dev_installed_versions:
                if version_to_set_up in version:
                    dev["version"] = version
                    break
            if not dev["version"]:
                logger.error(f"Software version {version_to_set_up} for {dev} is not included in available_versions")
        return devs

    def get_device_list_in_installed(self, version_to_set_up: str, devices: List[Device]) -> List[dict]:
        """
        Create version key for every device dict in device list, if requested version
        is in installed versions

        Args:
            version_to_set_up (str): requested version

        Returns:
            list : list of devices
        """
        return self._get_device_list_in(version_to_set_up, devices, "installed_versions")

    def get_device_list_in_available(self, version_to_set_up: str, devices: List[Device]) -> List[dict]:
        """
        Create version key for every device dict in device list, if requested version
        is in available versions

        Args:
            version_to_set_up (str): requested version

        Returns:
            list : list of devices
        """
        return self._get_device_list_in(version_to_set_up, devices, "available_versions")

    def get_device_list(self, devices: List[Device]) -> List[dict]:

        return [{"deviceId": dev.uuid, "deviceIP": dev.id} for dev in devices]



import unittest
from unittest.mock import patch

from parameterized import parameterized  # type: ignore

from vmngclient.api.administration import (
    AdministrationSettingsAPI,
    ClusterManagementAPI,
    UserAlreadyExistsError,
    UserDoesNotExists,
    UsersAPI,
)
from vmngclient.dataclasses import CloudConnectorData, CloudServicesSettings, ServiceConfigurationData, User
from vmngclient.utils.creation_tools import create_dataclass


class TestUsersAPI(unittest.TestCase):
    def setUp(self) -> None:
        self.users = [
            {"userName": "admin", "locale": "en_US", "group": []},
            {
                "userName": "demo_user",
                "password": "qwer",
                "description": "Demo User",
                "resGroupName": "group1",
                "locale": "en_US",
                "group": ["basic"],
            },
        ]
        self.user_dataclass = [create_dataclass(User, user) for user in self.users]

    @patch("vmngclient.session.Session")
    def test_get_all_users(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.users
        # Act
        answer = UsersAPI(mock_session).get_all_users()
        # Assert
        self.assertEqual(answer, self.user_dataclass)

    @patch("vmngclient.session.Session")
    def test_exists_true(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.users
        # Act
        answer = UsersAPI(mock_session).exists("demo_user")
        # Assert
        self.assertTrue(answer)

    @patch("vmngclient.session.Session")
    def test_exists_false(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.users
        # Act
        answer = UsersAPI(mock_session).exists("no_user")
        # Assert
        self.assertFalse(answer)

    @parameterized.expand([[200, True], [400, False]])
    @patch("vmngclient.session.Session")
    @patch("requests.Response")
    def test_create_user(self, status_code, expected_outcome, mock_session, mock_response):
        # Arrange
        mock_session.get_data.return_value = [self.users[1]]
        mock_session.post.return_value = mock_response
        mock_response.status_code = status_code
        # Act
        answer = UsersAPI(mock_session).create_user(self.user_dataclass[0])
        # Assert
        self.assertEqual(answer, expected_outcome)

    @patch("vmngclient.session.Session")
    def test_create_user_existing(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.users

        # Act
        def answer():
            UsersAPI(mock_session).create_user(self.user_dataclass[0])

        # Assert
        self.assertRaises(UserAlreadyExistsError, answer)

    @parameterized.expand([[200, True], [400, False]])
    @patch("vmngclient.session.Session")
    @patch("requests.Response")
    def test_delete_user(self, status_code, expected_outcome, mock_session, mock_response):
        # Arrange
        mock_session.get_data.return_value = self.users
        mock_session.delete.return_value = mock_response
        mock_response.status_code = status_code
        # Act
        answer = UsersAPI(mock_session).delete_user("admin")
        # Assert
        self.assertEqual(answer, expected_outcome)

    @patch("vmngclient.session.Session")
    def test_delete_user_not_existing(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = [self.users[1]]

        # Act
        def answer():
            UsersAPI(mock_session).delete_user("no_user")

        # Assert
        self.assertRaises(UserDoesNotExists, answer)


class TestClusterManagementAPI(unittest.TestCase):
    def setUp(self) -> None:
        self.service_configuration = {
            "vmanageID": "1.1.1.1",
            "deviceIP": "1.1.1.2",
            "services": "none",
        }
        self.service_configuration_dataclass = create_dataclass(ServiceConfigurationData, self.service_configuration)
        self.cluster_management_health_status = [
            {"deviceIP": "localhost"},
            {
                "statistics-db": True,
                "application-server": True,
                "messaging-server": True,
                "configuration-db": True,
                "deviceIP": "localhost",
            },
        ]

    @parameterized.expand([[200, True], [400, False]])
    @patch("vmngclient.session.Session")
    @patch("requests.Response")
    def test_modify_cluster_setup(self, status_code, expected_outcome, mock_session, mock_response):
        # Arrange
        mock_session.put.return_value = mock_response
        mock_response.status_code = status_code
        # Act
        answer = ClusterManagementAPI(mock_session).modify_cluster_setup(self.service_configuration_dataclass)
        # Assert
        self.assertEqual(answer, expected_outcome)

    @patch("vmngclient.session.Session")
    def test_get_cluster_management_health_status(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.cluster_management_health_status
        # Act
        answer = ClusterManagementAPI(mock_session).get_cluster_management_health_status()
        # Assert
        self.assertEqual(answer, self.cluster_management_health_status)


class TestAdministrationSettingsAPI(unittest.TestCase):
    def setUp(self) -> None:
        self.cloud_connector_data = {
            "clientId": "12345",
            "clientSecret": "secret",
            "orgName": "organization",
            "telemetryEnabled": True,
        }
        self.cloud_connector_data_dataclass = create_dataclass(CloudConnectorData, self.cloud_connector_data)
        self.cloud_services_settings = {"enabled": True}
        self.cloud_services_settings_dataclass = create_dataclass(CloudServicesSettings, self.cloud_services_settings)
        self.cloud_on_ramp = [{"": ""}]

    @patch("vmngclient.session.Session")
    def test_get_sdavc_cloud_connector_config(self, mock_session):
        # Arrange
        mock_session.get_json.return_value = self.cloud_connector_data
        # Act
        answer = AdministrationSettingsAPI(mock_session).get_sdavc_cloud_connector_config()
        # Assert
        self.assertEqual(answer, self.cloud_connector_data_dataclass)

    @parameterized.expand([[200, True], [400, False]])
    @patch("vmngclient.session.Session")
    @patch("requests.Response")
    def test_enable_sdavc_cloud_connector(self, status_code, expected_outcome, mock_session, mock_response):
        # Arrange
        mock_session.post.return_value = mock_response
        mock_response.status_code = status_code
        # Act
        answer = AdministrationSettingsAPI(mock_session).enable_sdavc_cloud_connector(
            self.cloud_connector_data_dataclass
        )
        # Assert
        self.assertEqual(answer, expected_outcome)

    @parameterized.expand([[200, True], [400, False]])
    @patch("vmngclient.session.Session")
    @patch("requests.Response")
    def test_disable_sdavc_cloud_connector(self, status_code, expected_outcome, mock_session, mock_response):
        # Arrange
        mock_session.put.return_value = mock_response
        mock_response.status_code = status_code
        # Act
        answer = AdministrationSettingsAPI(mock_session).disable_sdavc_cloud_connector()
        # Assert
        self.assertEqual(answer, expected_outcome)

    @patch("vmngclient.session.Session")
    def test_get_cloud_services(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = [self.cloud_services_settings]
        # Act
        answer = AdministrationSettingsAPI(mock_session).get_cloud_services()
        # Assert
        self.assertEqual(answer, self.cloud_services_settings_dataclass)

    @parameterized.expand([[200, True], [400, False]])
    @patch("vmngclient.session.Session")
    @patch("requests.Response")
    def test_set_cloud_services(self, status_code, expected_outcome, mock_session, mock_response):
        # Arrange
        mock_session.post.return_value = mock_response
        mock_response.status_code = status_code
        # Act
        answer = AdministrationSettingsAPI(mock_session).set_cloud_services(self.cloud_services_settings_dataclass)
        # Assert
        self.assertEqual(answer, expected_outcome)

    @patch("vmngclient.session.Session")
    def test_get_cloud_on_ramp_for_saas_mode(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.cloud_on_ramp
        # Act
        answer = AdministrationSettingsAPI(mock_session).get_cloud_on_ramp_for_saas_mode()
        # Assert
        self.assertEqual(answer, self.cloud_on_ramp)

    @parameterized.expand([[200, True], [400, False]])
    @patch("vmngclient.session.Session")
    @patch("requests.Response")
    def test_enable_cloud_on_ramp_for_saas_mode(self, status_code, expected_outcome, mock_session, mock_response):
        # Arrange
        mock_session.put.return_value = mock_response
        mock_response.status_code = status_code
        # Act
        answer = AdministrationSettingsAPI(mock_session).enable_cloud_on_ramp_for_saas_mode(True)
        # Assert
        self.assertEqual(answer, expected_outcome)



import unittest
from unittest.mock import patch

from vmngclient.api.task_status_api import TaskStatus, wait_for_completed


class TestTaskStatusApi(unittest.TestCase):
    def setUp(self):
        self.task = TaskStatus("Success", "success", [])

    @patch("vmngclient.session.vManageSession")
    def test_wait_for_completed_success(self, mock_session):

        # Prepare mock data
        mock_session.get_data.return_value = [{"status": "Success", "statusId": "success", "activity": []}]

        # Assert
        answer = wait_for_completed(mock_session, "mock_action_id", 3000, 5)
        self.assertEqual(answer, self.task, "job status incorrect")

    @patch("vmngclient.session.vManageSession")
    def test_wait_for_completed_status_out_of_range(self, mock_session):

        # Prepare mock data
        mock_session.get_data.return_value = [{"status": "Other_status", "statusId": "other_status", "activity": []}]

        # Assert
        answer = wait_for_completed(mock_session, "mock_action_id", 1, 1)
        self.assertEqual(answer, None, "job status incorrect")
